#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const fslib_1 = require("@yarnpkg/fslib");
const clipanion_1 = require("clipanion");
const globby_1 = tslib_1.__importDefault(require("globby"));
const micromatch_1 = tslib_1.__importDefault(require("micromatch"));
const module_1 = require("module");
const ts = tslib_1.__importStar(require("typescript"));
const ast = tslib_1.__importStar(require("./ast"));
const BUILTINS = new Set([
    ...(module_1.Module.builtinModules || []),
    ...(module_1.Module.builtinModules || []).map(mod => `node:${mod}`),
    `pnpapi`,
]);
function probablyMinified(content) {
    if (content.length > 1024 * 1024)
        return true;
    return false;
}
async function findFiles(pattern, cwd, { ignoredFolders = [] } = {}) {
    const files = await (0, globby_1.default)(pattern, {
        absolute: true,
        cwd: fslib_1.npath.fromPortablePath(cwd),
        ignore: [`**/node_modules/**`, ...ignoredFolders.map(p => `${fslib_1.npath.fromPortablePath(p)}/**`)],
        gitignore: true,
    });
    return files.map(p => {
        return fslib_1.npath.toPortablePath(p);
    });
}
async function parseFile(p) {
    const content = await fslib_1.xfs.readFilePromise(p, `utf8`);
    if (probablyMinified(content))
        return null;
    return await ts.createSourceFile(fslib_1.npath.fromPortablePath(p), content, ts.ScriptTarget.ES2015, 
    /*setParentNodes */ true);
}
function extractIdents(name) {
    // We also support webpack loaders
    const parts = name.split(/!/);
    const idents = [];
    for (const part of parts) {
        // Webpack loaders can have query strings
        const partWithQs = part.replace(/\?.*/, ``);
        const match = partWithQs.match(/^(?!\.{0,2}(\/|$))(@[^/]*\/)?([^/]+)/);
        if (!match)
            continue;
        const ident = core_1.structUtils.tryParseIdent(match[0]);
        if (!ident)
            continue;
        idents.push(ident);
    }
    return idents;
}
function isValidDependency(ident, { workspace }) {
    if (ident.identHash === workspace.anchoredLocator.identHash)
        return true;
    if (workspace.manifest.hasDependency(ident))
        return true;
    return false;
}
function checkForUndeclaredDependency(workspace, referenceNode, moduleName, { configuration, report }) {
    if (BUILTINS.has(moduleName))
        return;
    const idents = extractIdents(moduleName);
    for (const ident of idents) {
        if (isValidDependency(ident, { workspace }))
            continue;
        const prettyLocation = ast.prettyNodeLocation(configuration, referenceNode);
        report.reportError(core_1.MessageName.UNNAMED, `${prettyLocation}: Undeclared dependency on ${core_1.structUtils.prettyIdent(configuration, ident)}`);
    }
}
function checkForUnsafeWebpackLoaderAccess(workspace, initializerNode, { configuration, report }) {
    if (workspace.manifest.private)
        return;
    if (initializerNode.kind === ts.SyntaxKind.CallExpression && ast.getCallType(initializerNode) !== null)
        return;
    const prettyLocation = ast.prettyNodeLocation(configuration, initializerNode);
    report.reportWarning(core_1.MessageName.UNNAMED, `${prettyLocation}: Webpack configs from non-private packages should avoid referencing loaders without require.resolve`);
}
function checkForNodeModuleStrings(stringishNode, { configuration, report }) {
    const match = /node_modules(?!(\\{2}|\/)\.cache)/g.test(stringishNode.getText());
    if (match) {
        const prettyLocation = ast.prettyNodeLocation(configuration, stringishNode);
        report.reportWarning(core_1.MessageName.UNNAMED, `${prettyLocation}: Strings should avoid referencing the node_modules directory (prefer require.resolve)`);
    }
}
function processFile(workspace, file, { configuration, report }) {
    const importedModules = new Set();
    const processNode = (node) => {
        switch (node.kind) {
            case ts.SyntaxKind.ImportDeclaration:
                {
                    const decl = node;
                    const importTarget = ast.extractStaticString(decl.moduleSpecifier);
                    checkForUndeclaredDependency(workspace, decl, importTarget, { configuration, report });
                }
                break;
            case ts.SyntaxKind.ExportDeclaration:
                {
                    const decl = node;
                    if (!decl.moduleSpecifier)
                        break;
                    const importTarget = ast.extractStaticString(decl.moduleSpecifier);
                    checkForUndeclaredDependency(workspace, decl, importTarget, { configuration, report });
                }
                break;
            case ts.SyntaxKind.CallExpression:
                {
                    const call = node;
                    const callType = ast.getCallType(call);
                    if (callType === null)
                        break;
                    const staticImport = ast.extractStaticString(call.arguments[0]);
                    if (staticImport === null)
                        break;
                    checkForUndeclaredDependency(workspace, call, staticImport, { configuration, report });
                }
                break;
            case ts.SyntaxKind.PropertyAssignment:
                {
                    const property = node;
                    const name = ast.extractStaticName(property.name);
                    if (name === null)
                        break;
                    if (name === `use` || name === `loader`) {
                        checkForUnsafeWebpackLoaderAccess(workspace, property.initializer, { configuration, report });
                    }
                }
                break;
            case ts.SyntaxKind.StringLiteral:
                {
                    const stringNode = node;
                    checkForNodeModuleStrings(stringNode, { configuration, report });
                }
                break;
            case ts.SyntaxKind.NoSubstitutionTemplateLiteral:
                {
                    const stringNode = node;
                    checkForNodeModuleStrings(stringNode, { configuration, report });
                }
                break;
            case ts.SyntaxKind.TemplateExpression:
                {
                    const stringNode = node;
                    checkForNodeModuleStrings(stringNode, { configuration, report });
                }
                break;
        }
        ts.forEachChild(node, processNode);
    };
    processNode(file);
    return importedModules;
}
async function buildJsonNode(p, accesses) {
    const content = await fslib_1.xfs.readFilePromise(p, `utf8`);
    // Just to be sure that it's valid JSON
    JSON.parse(content);
    const sourceFile = await ts.createSourceFile(fslib_1.npath.fromPortablePath(p), content, ts.ScriptTarget.JSON, 
    /*setParentNodes */ true);
    // @ts-expect-error
    let node = sourceFile.statements[0].expression;
    if (!node)
        throw new Error(`Invalid source tree`);
    for (let t = 0; t < accesses.length; ++t) {
        if (node.kind !== ts.SyntaxKind.ObjectLiteralExpression)
            throw new Error(`Invalid property type: ${ts.SyntaxKind[node.kind]}`);
        const literal = node;
        const property = literal.properties.find(property => {
            const name = property.name;
            return name.text === accesses[t];
        });
        if (!node)
            throw new Error(`Missing property "${accesses[t]}"`);
        node = property.initializer;
    }
    return node;
}
async function checkForUnmetPeerDependency(workspace, dependencyType, via, peer, { configuration, report }) {
    if (dependencyType === `dependencies` && workspace.manifest.hasConsumerDependency(peer))
        return;
    if (dependencyType === `devDependencies` && workspace.manifest.hasHardDependency(peer))
        return;
    if (workspace.manifest.name?.identHash === peer.identHash)
        return;
    const propertyNode = await buildJsonNode(fslib_1.ppath.join(workspace.cwd, core_1.Manifest.fileName), [dependencyType, core_1.structUtils.stringifyIdent(via)]);
    const prettyLocation = ast.prettyNodeLocation(configuration, propertyNode);
    report.reportError(core_1.MessageName.UNNAMED, `${prettyLocation}: Unmet transitive peer dependency on ${core_1.structUtils.prettyDescriptor(configuration, peer)}, via ${core_1.structUtils.prettyDescriptor(configuration, via)}`);
}
async function makeResolveFn(project) {
    const cache = await core_1.Cache.find(project.configuration);
    const resolver = project.configuration.makeResolver();
    const fetcher = project.configuration.makeFetcher();
    const checksums = project.storedChecksums;
    const yarnReport = new core_1.ThrowReport();
    const fetchOptions = { project, fetcher, cache, checksums, report: yarnReport, cacheOptions: { skipIntegrityCheck: true } };
    const resolveOptions = { ...fetchOptions, resolver, fetchOptions };
    return async (descriptor) => {
        const candidates = await resolver.getCandidates(descriptor, {}, resolveOptions);
        if (candidates.length === 0)
            return null;
        return await resolver.resolve(candidates[0], resolveOptions);
    };
}
async function processManifest(workspace, { configuration, report }) {
    const resolveFn = await makeResolveFn(workspace.project);
    for (const dependencyType of core_1.Manifest.hardDependencies) {
        for (const viaDescriptor of workspace.manifest[dependencyType].values()) {
            let pkg;
            try {
                pkg = await resolveFn(viaDescriptor);
            }
            catch (error) {
                report.reportWarning(core_1.MessageName.UNNAMED, `Resolving ${core_1.structUtils.prettyDescriptor(configuration, viaDescriptor)} errored with ${error.message}`);
                continue;
            }
            if (!pkg) {
                report.reportWarning(core_1.MessageName.UNNAMED, `Couldn't find a valid resolution for ${core_1.structUtils.prettyDescriptor(configuration, viaDescriptor)}`);
                continue;
            }
            for (const peerDescriptor of pkg.peerDependencies.values()) {
                // No need to check optional peer dependencies at all
                const peerDependencyMeta = pkg.peerDependenciesMeta.get(core_1.structUtils.stringifyIdent(peerDescriptor));
                if (typeof peerDependencyMeta !== `undefined` && peerDependencyMeta.optional)
                    continue;
                await checkForUnmetPeerDependency(workspace, dependencyType, viaDescriptor, peerDescriptor, { configuration, report });
            }
        }
    }
}
async function processWorkspace(workspace, { configuration, fileList, report }) {
    const progress = core_1.StreamReport.progressViaCounter(fileList.length + 1);
    const reportedProgress = report.reportProgress(progress);
    for (const scriptName of workspace.manifest.scripts.keys())
        if (scriptName.match(/^(pre|post)(?!(install|pack)$)/) && !scriptName.match(/^prettier/))
            report.reportWarning(core_1.MessageName.UNNAMED, `User scripts prefixed with "pre" or "post" (like "${scriptName}") will not be called in sequence anymore; prefer calling prologues and epilogues explicitly`);
    if (Array.isArray(workspace.manifest.raw.bundleDependencies) || Array.isArray(workspace.manifest.raw.bundledDependencies))
        report.reportWarning(core_1.MessageName.UNNAMED, `The bundleDependencies (or bundledDependencies) field is not supported anymore; prefer using a bundler`);
    for (const p of fileList) {
        const parsed = await parseFile(p);
        if (parsed !== null)
            processFile(workspace, parsed, { configuration, report });
        progress.tick();
    }
    await processManifest(workspace, { configuration, report });
    progress.tick();
    await reportedProgress;
}
class EntryCommand extends clipanion_1.Command {
    constructor() {
        super(...arguments);
        this.cwd = clipanion_1.Option.String({ required: false });
    }
    async execute() {
        const cwd = fslib_1.npath.toPortablePath(fslib_1.npath.resolve(this.cwd ?? `.`));
        const configuration = await core_1.Configuration.find(cwd, null, { strict: false });
        const allManifests = await findFiles(`**/package.json`, cwd);
        const allManifestFolders = allManifests.map(p => fslib_1.ppath.dirname(p));
        const allFiles = await findFiles(`**/*.{ts,tsx,js,jsx}`, cwd);
        const pluginConfiguration = (0, cli_1.getPluginConfiguration)();
        const findWorkspace = async (manifestCwd) => {
            const configuration = await core_1.Configuration.find(manifestCwd, pluginConfiguration, { strict: false });
            if (!configuration.projectCwd)
                return null;
            const { project } = await core_1.Project.find(configuration, configuration.projectCwd);
            const workspace = project.tryWorkspaceByCwd(manifestCwd);
            if (!workspace)
                return null;
            return workspace;
        };
        const report = await core_1.StreamReport.start({
            configuration,
            stdout: this.context.stdout,
        }, async (report) => {
            report.reportInfo(core_1.MessageName.UNNAMED, `Found ${allManifestFolders.length} package(s) to process`);
            report.reportInfo(core_1.MessageName.UNNAMED, `For a grand total of ${allFiles.length} file(s) to validate`);
            const progress = core_1.StreamReport.progressViaCounter(allManifestFolders.length);
            const reportedProgress = report.reportProgress(progress);
            try {
                for (const manifestFolder of allManifestFolders) {
                    const manifestPath = fslib_1.ppath.join(manifestFolder, core_1.Manifest.fileName);
                    report.reportSeparator();
                    try {
                        await report.startTimerPromise(manifestPath, async () => {
                            const workspace = await findWorkspace(manifestFolder);
                            if (!workspace)
                                return;
                            const patterns = [`${manifestFolder}/**`];
                            const ignore = [];
                            for (const otherManifestFolder of allManifestFolders) {
                                const sub = fslib_1.ppath.contains(manifestFolder, otherManifestFolder);
                                if (sub !== null && sub !== `.`) {
                                    ignore.push(`${otherManifestFolder}/**`);
                                }
                            }
                            const fileList = (0, micromatch_1.default)(allFiles, patterns, { ignore });
                            await processWorkspace(workspace, {
                                configuration,
                                fileList,
                                report,
                            });
                        });
                    }
                    catch { }
                    progress.tick();
                }
            }
            catch (error) {
                reportedProgress.stop();
                throw error;
            }
            report.reportSeparator();
            await reportedProgress;
        });
        return report.exitCode();
    }
}
const cli = new clipanion_1.Cli({
    binaryLabel: `Yarn Doctor`,
    binaryName: `yarn dlx @yarnpkg/doctor`,
    binaryVersion: require(`@yarnpkg/doctor/package.json`).version,
});
cli.register(EntryCommand);
cli.register(clipanion_1.Builtins.VersionCommand);
cli.runExit(process.argv.slice(2), {
    stdin: process.stdin,
    stdout: process.stdout,
    stderr: process.stderr,
});
