"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const clipanion_1 = require("clipanion");
const lodash_clonedeep_1 = tslib_1.__importDefault(require("lodash.clonedeep"));
const lodash_has_1 = tslib_1.__importDefault(require("lodash.has"));
const lodash_unset_1 = tslib_1.__importDefault(require("lodash.unset"));
// eslint-disable-next-line arca/no-default-export
class ConfigUnsetCommand extends cli_1.BaseCommand {
    constructor() {
        super(...arguments);
        this.home = clipanion_1.Option.Boolean(`-H,--home`, false, {
            description: `Update the home configuration instead of the project configuration`,
        });
        this.name = clipanion_1.Option.String();
    }
    async execute() {
        const configuration = await core_1.Configuration.find(this.context.cwd, this.context.plugins);
        const assertProjectCwd = () => {
            if (!configuration.projectCwd)
                throw new clipanion_1.UsageError(`This command must be run from within a project folder`);
            return configuration.projectCwd;
        };
        const name = this.name.replace(/[.[].*$/, ``);
        const path = this.name.replace(/^[^.[]*\.?/, ``);
        const setting = configuration.settings.get(name);
        if (typeof setting === `undefined`)
            throw new clipanion_1.UsageError(`Couldn't find a configuration settings named "${name}"`);
        const updateConfiguration = this.home
            ? patch => core_1.Configuration.updateHomeConfiguration(patch)
            : patch => core_1.Configuration.updateConfiguration(assertProjectCwd(), patch);
        const report = await core_1.StreamReport.start({
            configuration,
            includeFooter: false,
            stdout: this.context.stdout,
        }, async (report) => {
            let bailedOutEarly = false;
            await updateConfiguration(current => {
                if (!(0, lodash_has_1.default)(current, this.name)) {
                    report.reportWarning(core_1.MessageName.UNNAMED, `Configuration doesn't contain setting ${this.name}; there is nothing to unset`);
                    bailedOutEarly = true;
                    return current;
                }
                const clone = path
                    ? (0, lodash_clonedeep_1.default)(current)
                    : { ...current };
                (0, lodash_unset_1.default)(clone, this.name);
                return clone;
            });
            // We can't show the success message in the callback as we must first wait for the new configuration to be persisted
            if (!bailedOutEarly) {
                report.reportInfo(core_1.MessageName.UNNAMED, `Successfully unset ${this.name}`);
            }
        });
        return report.exitCode();
    }
}
exports.default = ConfigUnsetCommand;
ConfigUnsetCommand.paths = [
    [`config`, `unset`],
];
ConfigUnsetCommand.usage = clipanion_1.Command.Usage({
    description: `unset a configuration setting`,
    details: `
      This command will unset a configuration setting.
    `,
    examples: [[
            `Unset a simple configuration setting`,
            `yarn config unset initScope`,
        ], [
            `Unset a complex configuration setting`,
            `yarn config unset packageExtensions`,
        ], [
            `Unset a nested configuration setting`,
            `yarn config unset npmScopes.company.npmRegistryServer`,
        ]],
});
