"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.del = exports.put = exports.post = exports.get = exports.getPackageMetadata = exports.getIdentUrl = exports.customPackageError = exports.handleInvalidAuthenticationError = exports.AuthType = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@yarnpkg/core");
const core_2 = require("@yarnpkg/core");
const fslib_1 = require("@yarnpkg/fslib");
const enquirer_1 = require("enquirer");
const lodash_pick_1 = tslib_1.__importDefault(require("lodash.pick"));
const semver_1 = tslib_1.__importDefault(require("semver"));
const url_1 = require("url");
const npmConfigUtils = tslib_1.__importStar(require("./npmConfigUtils"));
var AuthType;
(function (AuthType) {
    AuthType[AuthType["NO_AUTH"] = 0] = "NO_AUTH";
    AuthType[AuthType["BEST_EFFORT"] = 1] = "BEST_EFFORT";
    AuthType[AuthType["CONFIGURATION"] = 2] = "CONFIGURATION";
    AuthType[AuthType["ALWAYS_AUTH"] = 3] = "ALWAYS_AUTH";
})(AuthType = exports.AuthType || (exports.AuthType = {}));
/**
 * Consumes all 401 Unauthorized errors and reports them as `AUTHENTICATION_INVALID`.
 *
 * It doesn't handle 403 Forbidden, as the npm registry uses it when the user attempts
 * a prohibited action, such as publishing a package with a similar name to an existing package.
 */
async function handleInvalidAuthenticationError(error, { attemptedAs, registry, headers, configuration }) {
    if (isOtpError(error))
        throw new core_2.ReportError(core_2.MessageName.AUTHENTICATION_INVALID, `Invalid OTP token`);
    if (error.originalError?.name === `HTTPError` && error.originalError?.response.statusCode === 401) {
        throw new core_2.ReportError(core_2.MessageName.AUTHENTICATION_INVALID, `Invalid authentication (${typeof attemptedAs !== `string` ? `as ${await whoami(registry, headers, { configuration })}` : `attempted as ${attemptedAs}`})`);
    }
}
exports.handleInvalidAuthenticationError = handleInvalidAuthenticationError;
function customPackageError(error, configuration) {
    const statusCode = error.response?.statusCode;
    if (!statusCode)
        return null;
    if (statusCode === 404)
        return `Package not found`;
    if (statusCode >= 500 && statusCode < 600)
        return `The registry appears to be down (using a ${core_1.formatUtils.applyHyperlink(configuration, `local cache`, `https://yarnpkg.com/advanced/lexicon#local-cache`)} might have protected you against such outages)`;
    return null;
}
exports.customPackageError = customPackageError;
function getIdentUrl(ident) {
    if (ident.scope) {
        return `/@${ident.scope}%2f${ident.name}`;
    }
    else {
        return `/${ident.name}`;
    }
}
exports.getIdentUrl = getIdentUrl;
// We use 2 different caches:
// - an in-memory cache, to avoid hitting the disk and the network more than once per process for each package
// - an on-disk cache, for exact version matches and to avoid refetching the metadata if the resource hasn't changed on the server
const PACKAGE_METADATA_CACHE = new Map();
/**
 * Caches and returns the package metadata for the given ident.
 *
 * Note: This function only caches and returns specific fields from the metadata.
 * If you need other fields, use the uncached {@link get} or consider whether it would make more sense to extract
 * the fields from the on-disk packages using the linkers or from the fetch results using the fetchers.
 */
async function getPackageMetadata(ident, { cache, project, registry, headers, version, ...rest }) {
    return await core_1.miscUtils.getFactoryWithDefault(PACKAGE_METADATA_CACHE, ident.identHash, async () => {
        const { configuration } = project;
        registry = normalizeRegistry(configuration, { ident, registry });
        const registryFolder = getRegistryFolder(configuration, registry);
        const identPath = fslib_1.ppath.join(registryFolder, `${core_1.structUtils.slugifyIdent(ident)}.json`);
        let cached = null;
        // We bypass the on-disk cache for security reasons if the lockfile needs to be refreshed,
        // since most likely the user is trying to validate the metadata using hardened mode.
        if (!project.lockfileNeedsRefresh) {
            try {
                cached = await fslib_1.xfs.readJsonPromise(identPath);
            }
            catch { }
            if (cached) {
                if (typeof version !== `undefined` && typeof cached.metadata.versions[version] !== `undefined`)
                    return cached.metadata;
                if (configuration.get(`enableOfflineMode`)) {
                    const copy = structuredClone(cached.metadata);
                    const deleted = new Set();
                    if (cache) {
                        for (const version of Object.keys(copy.versions)) {
                            const locator = core_1.structUtils.makeLocator(ident, `npm:${version}`);
                            const mirrorPath = cache.getLocatorMirrorPath(locator);
                            if (!mirrorPath || !fslib_1.xfs.existsSync(mirrorPath)) {
                                delete copy.versions[version];
                                deleted.add(version);
                            }
                        }
                        const latest = copy[`dist-tags`].latest;
                        if (deleted.has(latest)) {
                            const allVersions = Object.keys(cached.metadata.versions)
                                .sort(semver_1.default.compare);
                            let latestIndex = allVersions.indexOf(latest);
                            while (deleted.has(allVersions[latestIndex]) && latestIndex >= 0)
                                latestIndex -= 1;
                            if (latestIndex >= 0) {
                                copy[`dist-tags`].latest = allVersions[latestIndex];
                            }
                            else {
                                delete copy[`dist-tags`].latest;
                            }
                        }
                    }
                    return copy;
                }
            }
        }
        return await get(getIdentUrl(ident), {
            ...rest,
            customErrorMessage: customPackageError,
            configuration,
            registry,
            ident,
            headers: {
                ...headers,
                // We set both headers in case a registry doesn't support ETags
                [`If-None-Match`]: cached?.etag,
                [`If-Modified-Since`]: cached?.lastModified,
            },
            wrapNetworkRequest: async (executor) => async () => {
                const response = await executor();
                if (response.statusCode === 304) {
                    if (cached === null)
                        throw new Error(`Assertion failed: cachedMetadata should not be null`);
                    return {
                        ...response,
                        body: cached.metadata,
                    };
                }
                const packageMetadata = pickPackageMetadata(JSON.parse(response.body.toString()));
                PACKAGE_METADATA_CACHE.set(ident.identHash, packageMetadata);
                const metadata = {
                    metadata: packageMetadata,
                    etag: response.headers.etag,
                    lastModified: response.headers[`last-modified`],
                };
                // We append the PID because it is guaranteed that this code is only run once per process for a given ident
                const identPathTemp = `${identPath}-${process.pid}.tmp`;
                await fslib_1.xfs.mkdirPromise(registryFolder, { recursive: true });
                await fslib_1.xfs.writeJsonPromise(identPathTemp, metadata, { compact: true });
                // Doing a rename is important to ensure the cache is atomic
                await fslib_1.xfs.renamePromise(identPathTemp, identPath);
                return {
                    ...response,
                    body: packageMetadata,
                };
            },
        });
    });
}
exports.getPackageMetadata = getPackageMetadata;
const CACHED_FIELDS = [
    `name`,
    `dist.tarball`,
    `bin`,
    `scripts`,
    `os`,
    `cpu`,
    `libc`,
    `dependencies`,
    `dependenciesMeta`,
    `optionalDependencies`,
    `peerDependencies`,
    `peerDependenciesMeta`,
    `deprecated`,
];
function pickPackageMetadata(metadata) {
    return {
        'dist-tags': metadata[`dist-tags`],
        versions: Object.fromEntries(Object.entries(metadata.versions).map(([key, value]) => [
            key,
            (0, lodash_pick_1.default)(value, CACHED_FIELDS),
        ])),
    };
}
/**
 * Used to invalidate the on-disk cache when the format changes.
 */
const CACHE_KEY = core_1.hashUtils.makeHash(...CACHED_FIELDS).slice(0, 6);
function getRegistryFolder(configuration, registry) {
    const metadataFolder = getMetadataFolder(configuration);
    const parsed = new url_1.URL(registry);
    return fslib_1.ppath.join(metadataFolder, CACHE_KEY, parsed.hostname);
}
function getMetadataFolder(configuration) {
    return fslib_1.ppath.join(configuration.get(`globalFolder`), `metadata/npm`);
}
async function get(path, { configuration, headers, ident, authType, registry, ...rest }) {
    registry = normalizeRegistry(configuration, { ident, registry });
    if (ident && ident.scope && typeof authType === `undefined`)
        authType = AuthType.BEST_EFFORT;
    const auth = await getAuthenticationHeader(registry, { authType, configuration, ident });
    if (auth)
        headers = { ...headers, authorization: auth };
    try {
        return await core_1.httpUtils.get(path.charAt(0) === `/` ? `${registry}${path}` : path, { configuration, headers, ...rest });
    }
    catch (error) {
        await handleInvalidAuthenticationError(error, { registry, configuration, headers });
        throw error;
    }
}
exports.get = get;
async function post(path, body, { attemptedAs, configuration, headers, ident, authType = AuthType.ALWAYS_AUTH, registry, otp, ...rest }) {
    registry = normalizeRegistry(configuration, { ident, registry });
    const auth = await getAuthenticationHeader(registry, { authType, configuration, ident });
    if (auth)
        headers = { ...headers, authorization: auth };
    if (otp)
        headers = { ...headers, ...getOtpHeaders(otp) };
    try {
        return await core_1.httpUtils.post(registry + path, body, { configuration, headers, ...rest });
    }
    catch (error) {
        if (!isOtpError(error) || otp) {
            await handleInvalidAuthenticationError(error, { attemptedAs, registry, configuration, headers });
            throw error;
        }
        otp = await askForOtp(error, { configuration });
        const headersWithOtp = { ...headers, ...getOtpHeaders(otp) };
        // Retrying request with OTP
        try {
            return await core_1.httpUtils.post(`${registry}${path}`, body, { configuration, headers: headersWithOtp, ...rest });
        }
        catch (error) {
            await handleInvalidAuthenticationError(error, { attemptedAs, registry, configuration, headers });
            throw error;
        }
    }
}
exports.post = post;
async function put(path, body, { attemptedAs, configuration, headers, ident, authType = AuthType.ALWAYS_AUTH, registry, otp, ...rest }) {
    registry = normalizeRegistry(configuration, { ident, registry });
    const auth = await getAuthenticationHeader(registry, { authType, configuration, ident });
    if (auth)
        headers = { ...headers, authorization: auth };
    if (otp)
        headers = { ...headers, ...getOtpHeaders(otp) };
    try {
        return await core_1.httpUtils.put(registry + path, body, { configuration, headers, ...rest });
    }
    catch (error) {
        if (!isOtpError(error)) {
            await handleInvalidAuthenticationError(error, { attemptedAs, registry, configuration, headers });
            throw error;
        }
        otp = await askForOtp(error, { configuration });
        const headersWithOtp = { ...headers, ...getOtpHeaders(otp) };
        // Retrying request with OTP
        try {
            return await core_1.httpUtils.put(`${registry}${path}`, body, { configuration, headers: headersWithOtp, ...rest });
        }
        catch (error) {
            await handleInvalidAuthenticationError(error, { attemptedAs, registry, configuration, headers });
            throw error;
        }
    }
}
exports.put = put;
async function del(path, { attemptedAs, configuration, headers, ident, authType = AuthType.ALWAYS_AUTH, registry, otp, ...rest }) {
    registry = normalizeRegistry(configuration, { ident, registry });
    const auth = await getAuthenticationHeader(registry, { authType, configuration, ident });
    if (auth)
        headers = { ...headers, authorization: auth };
    if (otp)
        headers = { ...headers, ...getOtpHeaders(otp) };
    try {
        return await core_1.httpUtils.del(registry + path, { configuration, headers, ...rest });
    }
    catch (error) {
        if (!isOtpError(error) || otp) {
            await handleInvalidAuthenticationError(error, { attemptedAs, registry, configuration, headers });
            throw error;
        }
        otp = await askForOtp(error, { configuration });
        const headersWithOtp = { ...headers, ...getOtpHeaders(otp) };
        // Retrying request with OTP
        try {
            return await core_1.httpUtils.del(`${registry}${path}`, { configuration, headers: headersWithOtp, ...rest });
        }
        catch (error) {
            await handleInvalidAuthenticationError(error, { attemptedAs, registry, configuration, headers });
            throw error;
        }
    }
}
exports.del = del;
function normalizeRegistry(configuration, { ident, registry }) {
    if (typeof registry === `undefined` && ident)
        return npmConfigUtils.getScopeRegistry(ident.scope, { configuration });
    if (typeof registry !== `string`)
        throw new Error(`Assertion failed: The registry should be a string`);
    return npmConfigUtils.normalizeRegistry(registry);
}
async function getAuthenticationHeader(registry, { authType = AuthType.CONFIGURATION, configuration, ident }) {
    const effectiveConfiguration = npmConfigUtils.getAuthConfiguration(registry, { configuration, ident });
    const mustAuthenticate = shouldAuthenticate(effectiveConfiguration, authType);
    if (!mustAuthenticate)
        return null;
    const header = await configuration.reduceHook((hooks) => {
        return hooks.getNpmAuthenticationHeader;
    }, undefined, registry, { configuration, ident });
    if (header)
        return header;
    if (effectiveConfiguration.get(`npmAuthToken`))
        return `Bearer ${effectiveConfiguration.get(`npmAuthToken`)}`;
    if (effectiveConfiguration.get(`npmAuthIdent`)) {
        const npmAuthIdent = effectiveConfiguration.get(`npmAuthIdent`);
        if (npmAuthIdent.includes(`:`))
            return `Basic ${Buffer.from(npmAuthIdent).toString(`base64`)}`;
        return `Basic ${npmAuthIdent}`;
    }
    if (mustAuthenticate && authType !== AuthType.BEST_EFFORT) {
        throw new core_2.ReportError(core_2.MessageName.AUTHENTICATION_NOT_FOUND, `No authentication configured for request`);
    }
    else {
        return null;
    }
}
function shouldAuthenticate(authConfiguration, authType) {
    switch (authType) {
        case AuthType.CONFIGURATION:
            return authConfiguration.get(`npmAlwaysAuth`);
        case AuthType.BEST_EFFORT:
        case AuthType.ALWAYS_AUTH:
            return true;
        case AuthType.NO_AUTH:
            return false;
        default:
            throw new Error(`Unreachable`);
    }
}
async function whoami(registry, headers, { configuration }) {
    if (typeof headers === `undefined` || typeof headers.authorization === `undefined`)
        return `an anonymous user`;
    try {
        const response = await core_1.httpUtils.get(new url_1.URL(`${registry}/-/whoami`).href, {
            configuration,
            headers,
            jsonResponse: true,
        });
        return response.username ?? `an unknown user`;
    }
    catch {
        return `an unknown user`;
    }
}
async function askForOtp(error, { configuration }) {
    const notice = error.originalError?.response.headers[`npm-notice`];
    if (notice) {
        await core_1.StreamReport.start({
            configuration,
            stdout: process.stdout,
            includeFooter: false,
        }, async (report) => {
            report.reportInfo(core_2.MessageName.UNNAMED, notice.replace(/(https?:\/\/\S+)/g, core_1.formatUtils.pretty(configuration, `$1`, core_1.formatUtils.Type.URL)));
            if (!process.env.YARN_IS_TEST_ENV) {
                const autoOpen = notice.match(/open (https?:\/\/\S+)/i);
                if (autoOpen && core_1.nodeUtils.openUrl) {
                    const { openNow } = await (0, enquirer_1.prompt)({
                        type: `confirm`,
                        name: `openNow`,
                        message: `Do you want to try to open this url now?`,
                        required: true,
                        initial: true,
                        onCancel: () => process.exit(130),
                    });
                    if (openNow) {
                        if (!await core_1.nodeUtils.openUrl(autoOpen[1])) {
                            report.reportSeparator();
                            report.reportWarning(core_2.MessageName.UNNAMED, `We failed to automatically open the url; you'll have to open it yourself in your browser of choice.`);
                        }
                    }
                }
            }
        });
        process.stdout.write(`\n`);
    }
    if (process.env.YARN_IS_TEST_ENV)
        return process.env.YARN_INJECT_NPM_2FA_TOKEN || ``;
    const { otp } = await (0, enquirer_1.prompt)({
        type: `password`,
        name: `otp`,
        message: `One-time password:`,
        required: true,
        onCancel: () => process.exit(130),
    });
    process.stdout.write(`\n`);
    return otp;
}
function isOtpError(error) {
    if (error.originalError?.name !== `HTTPError`)
        return false;
    try {
        const authMethods = error.originalError?.response.headers[`www-authenticate`].split(/,\s*/).map((s) => s.toLowerCase());
        return authMethods.includes(`otp`);
    }
    catch (e) {
        return false;
    }
}
function getOtpHeaders(otp) {
    return {
        [`npm-otp`]: otp,
    };
}
