"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Driver = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@yarnpkg/core");
const fslib_1 = require("@yarnpkg/fslib");
const stageUtils = tslib_1.__importStar(require("../stageUtils"));
const MESSAGE_MARKER = `Commit generated via \`yarn stage\``;
const COMMIT_DEPTH = 11;
async function getLastCommitHash(cwd) {
    const { code, stdout } = await core_1.execUtils.execvp(`git`, [`log`, `-1`, `--pretty=format:%H`], { cwd });
    if (code === 0) {
        return stdout.trim();
    }
    else {
        return null;
    }
}
async function genCommitMessage(cwd, changes) {
    const actions = [];
    const modifiedPkgJsonFiles = changes.filter(change => {
        return fslib_1.ppath.basename(change.path) === `package.json`;
    });
    for (const { action, path } of modifiedPkgJsonFiles) {
        const relativePath = fslib_1.ppath.relative(cwd, path);
        if (action === stageUtils.ActionType.MODIFY) {
            const commitHash = await getLastCommitHash(cwd);
            const { stdout: prevSource } = await core_1.execUtils.execvp(`git`, [`show`, `${commitHash}:${relativePath}`], { cwd, strict: true });
            const prevManifest = await core_1.Manifest.fromText(prevSource);
            const currManifest = await core_1.Manifest.fromFile(path);
            const allCurrDeps = new Map([...currManifest.dependencies, ...currManifest.devDependencies]);
            const allPrevDeps = new Map([...prevManifest.dependencies, ...prevManifest.devDependencies]);
            for (const [indentHash, value] of allPrevDeps) {
                const pkgName = core_1.structUtils.stringifyIdent(value);
                const currDep = allCurrDeps.get(indentHash);
                if (!currDep) {
                    actions.push([stageUtils.ActionType.REMOVE, pkgName]);
                }
                else if (currDep.range !== value.range) {
                    actions.push([stageUtils.ActionType.MODIFY, `${pkgName} to ${currDep.range}`]);
                }
            }
            for (const [indentHash, value] of allCurrDeps) {
                if (!allPrevDeps.has(indentHash)) {
                    actions.push([stageUtils.ActionType.ADD, core_1.structUtils.stringifyIdent(value)]);
                }
            }
        }
        else if (action === stageUtils.ActionType.CREATE) {
            // New package.json
            const manifest = await core_1.Manifest.fromFile(path);
            if (manifest.name) {
                actions.push([stageUtils.ActionType.CREATE, core_1.structUtils.stringifyIdent(manifest.name)]);
            }
            else {
                actions.push([stageUtils.ActionType.CREATE, `a package`]);
            }
        }
        else if (action === stageUtils.ActionType.DELETE) {
            const commitHash = await getLastCommitHash(cwd);
            const { stdout: prevSource } = await core_1.execUtils.execvp(`git`, [`show`, `${commitHash}:${relativePath}`], { cwd, strict: true });
            // Deleted package.json; we need to load it from its past sources
            const manifest = await core_1.Manifest.fromText(prevSource);
            if (manifest.name) {
                actions.push([stageUtils.ActionType.DELETE, core_1.structUtils.stringifyIdent(manifest.name)]);
            }
            else {
                actions.push([stageUtils.ActionType.DELETE, `a package`]);
            }
        }
        else {
            throw new Error(`Assertion failed: Unsupported action type`);
        }
    }
    const { code, stdout } = await core_1.execUtils.execvp(`git`, [`log`, `-${COMMIT_DEPTH}`, `--pretty=format:%s`], { cwd });
    const lines = code === 0
        ? stdout.split(/\n/g).filter((line) => line !== ``)
        : [];
    const consensus = stageUtils.findConsensus(lines);
    const message = stageUtils.genCommitMessage(consensus, actions);
    return message;
}
const unstagedPrefixes = {
    [stageUtils.ActionType.CREATE]: [` A `, `?? `],
    [stageUtils.ActionType.MODIFY]: [` M `],
    [stageUtils.ActionType.DELETE]: [` D `],
};
const stagedPrefixes = {
    [stageUtils.ActionType.CREATE]: [`A  `],
    [stageUtils.ActionType.MODIFY]: [`M  `],
    [stageUtils.ActionType.DELETE]: [`D  `],
};
exports.Driver = {
    async findRoot(cwd) {
        return await stageUtils.findVcsRoot(cwd, { marker: `.git` });
    },
    async filterChanges(cwd, yarnRoots, yarnNames, opts) {
        const { stdout } = await core_1.execUtils.execvp(`git`, [`status`, `-s`], { cwd, strict: true });
        const lines = stdout.toString().split(/\n/g);
        const changePrefix = opts?.staged ? stagedPrefixes : unstagedPrefixes;
        const changes = [].concat(...lines.map((line) => {
            if (line === ``)
                return [];
            const prefix = line.slice(0, 3);
            const path = fslib_1.ppath.resolve(cwd, line.slice(3));
            // New directories need to be expanded to their content
            if (!opts?.staged && prefix === `?? ` && line.endsWith(`/`)) {
                return stageUtils.expandDirectory(path).map(path => ({
                    action: stageUtils.ActionType.CREATE,
                    path,
                }));
            }
            else {
                const actions = [stageUtils.ActionType.CREATE, stageUtils.ActionType.MODIFY, stageUtils.ActionType.DELETE];
                const action = actions.find(action => changePrefix[action].includes(prefix));
                if (action !== undefined) {
                    return [{
                            action,
                            path,
                        }];
                }
                return [];
            }
        }));
        return changes.filter(change => {
            return stageUtils.isYarnFile(change.path, {
                roots: yarnRoots,
                names: yarnNames,
            });
        });
    },
    async genCommitMessage(cwd, changeList) {
        return await genCommitMessage(cwd, changeList);
    },
    async makeStage(cwd, changeList) {
        const localPaths = changeList.map(file => fslib_1.npath.fromPortablePath(file.path));
        await core_1.execUtils.execvp(`git`, [`add`, `--`, ...localPaths], { cwd, strict: true });
    },
    async makeCommit(cwd, changeList, commitMessage) {
        const localPaths = changeList.map(file => fslib_1.npath.fromPortablePath(file.path));
        await core_1.execUtils.execvp(`git`, [`add`, `-N`, `--`, ...localPaths], { cwd, strict: true });
        await core_1.execUtils.execvp(`git`, [`commit`, `-m`, `${commitMessage}\n\n${MESSAGE_MARKER}\n`, `--`, ...localPaths], { cwd, strict: true });
    },
    async makeReset(cwd, changeList) {
        const localPaths = changeList.map(path => fslib_1.npath.fromPortablePath(path.path));
        await core_1.execUtils.execvp(`git`, [`reset`, `HEAD`, `--`, ...localPaths], { cwd, strict: true });
    },
};
