"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("@yarnpkg/core");
const fslib_1 = require("@yarnpkg/fslib");
const clipanion_1 = require("clipanion");
const dynamicRequire_1 = require("../dynamicRequire");
const generateSdk_1 = require("../generateSdk");
// eslint-disable-next-line arca/no-default-export
class SdkCommand extends clipanion_1.Command {
    constructor() {
        super(...arguments);
        this.cwd = clipanion_1.Option.String(`--cwd`, process.cwd(), {
            description: `The directory to run the command in`,
        });
        this.verbose = clipanion_1.Option.Boolean(`-v,--verbose`, false, {
            description: `Print all skipped dependencies`,
        });
        this.integrations = clipanion_1.Option.Rest();
    }
    async execute() {
        let nextProjectRoot = fslib_1.npath.toPortablePath(this.cwd);
        let currProjectRoot = null;
        let pnpFilename;
        while (nextProjectRoot !== currProjectRoot) {
            currProjectRoot = nextProjectRoot;
            nextProjectRoot = fslib_1.ppath.dirname(currProjectRoot);
            if (fslib_1.xfs.existsSync(fslib_1.ppath.join(currProjectRoot, fslib_1.Filename.pnpCjs))) {
                pnpFilename = fslib_1.Filename.pnpCjs;
                break;
            }
            // TODO: Drop support for .pnp.js files after they stop being used.
            if (fslib_1.xfs.existsSync(fslib_1.ppath.join(currProjectRoot, fslib_1.Filename.pnpJs))) {
                pnpFilename = fslib_1.Filename.pnpJs;
                break;
            }
        }
        if (nextProjectRoot === currProjectRoot)
            throw new clipanion_1.UsageError(`This tool can only be used with projects using Yarn Plug'n'Play`);
        const configuration = core_1.Configuration.create(currProjectRoot);
        const pnpPath = fslib_1.ppath.join(currProjectRoot, pnpFilename);
        const pnpApi = (0, dynamicRequire_1.dynamicRequire)(fslib_1.npath.fromPortablePath(pnpPath));
        // Need to setup the fs patch so we can read from the archives
        pnpApi?.setup();
        const onlyBase = this.integrations.length === 1 && this.integrations[0] === `base`;
        const requestedIntegrations = this.integrations.length === 0 || onlyBase
            ? new Set()
            : new Set(this.integrations);
        (0, generateSdk_1.validateIntegrations)(requestedIntegrations);
        const topLevelInformation = pnpApi.getPackageInformation(pnpApi.topLevel);
        const projectRoot = fslib_1.npath.toPortablePath(topLevelInformation.packageLocation);
        const integrationsFile = await generateSdk_1.IntegrationsFile.find(projectRoot);
        const preexistingIntegrations = integrationsFile !== null
            ? integrationsFile.integrations
            : new Set();
        const allIntegrations = new Set([
            ...requestedIntegrations,
            ...preexistingIntegrations,
        ]);
        if (allIntegrations.size === 0 && !onlyBase)
            throw new clipanion_1.UsageError(`No integrations have been provided as arguments, and no preexisting integrations could be found. Make sure to run \`yarn sdks <integrations>\` first, or \`yarn sdks base\` if you only need the SDK files and prefer to manage your own environment settings. Run \`yarn sdks -h\` to see the list of supported integrations.`);
        const report = await core_1.StreamReport.start({
            configuration,
            includeFooter: false,
            stdout: this.context.stdout,
        }, async (report) => {
            await (0, generateSdk_1.generateSdk)(pnpApi, {
                requestedIntegrations,
                preexistingIntegrations,
            }, {
                report,
                onlyBase,
                configuration,
                verbose: this.verbose,
            });
        });
        return report.exitCode();
    }
}
exports.default = SdkCommand;
SdkCommand.paths = [
    clipanion_1.Command.Default,
];
SdkCommand.usage = clipanion_1.Command.Usage({
    description: `generate editor SDKs and settings`,
    details: `
      This command generates a new directory, \`.yarn/sdks\`, which includes the base SDKs.

      When used without arguments, it:

      - throws an error on non-pnpified projects

      - updates all existing SDKs and editor settings on already-pnpified projects

      The optional integrations arguments are a set of supported integrations, or the keyword \`base\`.

      - When \`base\` is used, only the base SDKs will be generated. This is useful for when an editor is not yet supported and you plan to manage the settings yourself.

      - When a set of integrations is used (e.g. \`vscode\`, \`vim\`, ...), the base SDKs will be installed plus all the settings relevant to the corresponding environments (for example on VSCode it would set \`typescript.tsdk\`).

      The supported integrations at this time are: ${[...generateSdk_1.SUPPORTED_INTEGRATIONS.keys()].map(integration => `\`${integration}\``).join(`, `)}.

      **Note:** This command always updates the already-installed SDKs and editor settings, no matter which arguments are passed.

      For more details on Editor SDKs, please consult the dedicated page from our website: https://yarnpkg.com/getting-started/editor-sdks.
    `,
    examples: [[
            `Generate the base SDKs`,
            `$0 base`,
        ], [
            `Generate the base SDKs and editor settings for supported editors`,
            `$0 vscode vim`,
        ], [
            `Update all generated SDKs and editor settings`,
            `$0`,
        ]],
});
